##' @title Fast Control Defaults
##' @return A Control object with parameters set.
##' @author Rich FitzJohn
##' @export
##' @param base An optional \code{Control} object.  If omitted, the
##' defaults are used.
fast_control <- function(base=Control()) {
  base$environment_light_rescale_usually <- TRUE
  base$environment_light_tol <- 1e-4

  base$plant_assimilation_adaptive <- FALSE
  base$plant_assimilation_rule <- 21
  base$plant_assimilation_over_distribution <- FALSE
  base$plant_assimilation_tol <- 1e-4

  base$ode_tol_rel <- 1e-4
  base$ode_tol_abs <- 1e-4
  base$ode_step_size_max <- 5

  base$cohort_gradient_direction <- -1
  base$cohort_gradient_richardson <- FALSE

  base
}

##' Control parameters for \code{\link{equilibrium_seed_rain}} that
##' make progress noisier.  This is just a convenience function.
##'
##' @title Noisy Parameters for Equilibrium Finding
##' @export
##' @param base An optional \code{Control} object.  If omitted, the
##' defaults are used.
equilibrium_verbose <- function(base=Control()) {
  base$schedule_verbose <- TRUE
  base$equilibrium_verbose <- TRUE
  base
}
##' @export
##' @rdname equilibrium_verbose
equilibrium_quiet <- function(base=Control()) {
  base$schedule_verbose <- FALSE
  base$equilibrium_verbose <- FALSE
  base
}

##' Run the SCM, returning the SCM object for interrogation
##'
##' This is the simplest way of using the SCM, probably.
##' @title Run SCM
##' @param p Parameters object
##' @param use_ode_times Should ODE times be used?
##' @return A \code{SCM} object.
##' @author Rich FitzJohn
##' @export
run_scm <- function(p, use_ode_times=FALSE) {
  type <- extract_RcppR6_template_type(p, "Parameters")
  scm <- SCM(type)(p)
  if (use_ode_times) {
    scm$use_ode_times <- TRUE
  }
  scm$run()
  scm
}

##' Hopefully sensible set of parameters for use with the SCM.  Turns
##' accuracy down a bunch, makes it noisy, sets up the
##' hyperparameterisation that we most often use.
##' @title Sensible, fast (ish) SCM parameters
##' @author Rich FitzJohn
##' @param type Name of model (defaults to FF16 but FF16r also valid)
##' @export
scm_base_parameters <- function(type="FF16") {
  ctrl <- equilibrium_verbose(fast_control())
  ctrl$schedule_eps <- 0.005
  ctrl$equilibrium_eps <- 1e-3
  Parameters(type)(patch_area=1.0, control=ctrl, hyperpar=hyperpar(type))
}

##' Run the SCM model, given a Parameters and CohortSchedule
##'
##' This is mostly a simple wrapper around some of the SCM functions.
##' Not sure if this is how we will generally want to do this.
##' Consider this function liable to change.
##'
##' @title Run the SCM, Collecting Output
##' @param p A \code{Parameters} object
##' @param include_area_leaf Include total leaf area (will change; see
##' issue #138)
##' @author Rich FitzJohn
##' @export
run_scm_collect <- function(p, include_area_leaf=FALSE) {
  collect_default <- function(scm) {
    scm$state
  }
  collect_area_leaf <- function(scm) {
    ret <- scm$state
    area_leaf <- numeric(length(ret$species))
    for (i in seq_along(ret$species)) {
      ## ret$species[[i]] <- rbind(
      ##   ret$species[[i]]
      ##   area_leaf=c(scm$patch$species[[i]]$area_leafs, 0.0))
      area_leaf[i] <- scm$patch$species[[i]]$area_leaf_above(0.0)
    }
    ret$area_leaf <- area_leaf
    ret
  }
  collect <- if (include_area_leaf) collect_area_leaf else collect_default
  type <- extract_RcppR6_template_type(p, "Parameters")

  scm <- SCM(type)(p)
  res <- list(collect(scm))

  while (!scm$complete) {
    scm$run_next()
    res <- c(res, list(collect(scm)))
  }

  time <- sapply(res, "[[", "time")
  light_env <- lapply(res, "[[", "light_env")
  species <- lapply(res, "[[", "species")
  ## The aperm() here means that dimensions are
  ## [variable,time,cohort], so that taking species[[1]]["height",,]
  ## gives a matrix that has time down rows and cohorts across columns
  ## (so is therefore plottable with matplot)
  species <- lapply(seq_along(species[[1]]), function(i)
                    aperm(pad_list_to_array(lapply(species, "[[", i)),
                          c(1, 3, 2)))
  ## Drop the boundary condition; we do this mostly because it cannot
  ## be compared against the reference output, which does not contain
  ## this.  This does have the nice property of giving a non-square
  ## matrix, so the difference between time and cohort becomes a
  ## little more obvious.
  species <- lapply(species, function(m) m[,,-dim(m)[[3]]])

  patch_density <- scm$patch$environment$disturbance_regime$density(time)

  ret <- list(time=time, species=species,
              light_env=light_env,
              seed_rain=scm$seed_rains,
              patch_density=patch_density,
              p=p)

  if (include_area_leaf) {
    ret$area_leaf <- do.call("rbind", lapply(res, "[[", "area_leaf"))
  }

  ret
}

##' Functions for reconstructing a Patch from an SCM
##' @title Reconstruct a patch
##' @param state State object created by \code{scm_state}
##' @param p Parameters object
##' @export
make_patch <- function(state, p) {
  type <- extract_RcppR6_template_type(p, "Parameters")
  n <- viapply(state$species, ncol)
  patch <- Patch(type)(p)
  patch$set_state(state$time, unlist(state$species), n, state$light_env)
  patch
}

##' @rdname make_patch
##' @param i Index to extract from \code{x}
##' @param x Result of running \code{\link{run_scm_collect}}
##' @export
scm_state <- function(i, x) {
  f_sp <- function(el) {
    el <- el[, i, ]
    el[, !is.na(el[1, ]), drop=FALSE]
  }
  list(time=x$time[[i]], species=lapply(x$species, f_sp),
       light_env=x$light_env[[i]])
}

##' @export
##' @rdname make_patch
scm_patch <- function(i, x) {
  make_patch(scm_state(i, x), x$p)
}

run_scm_error <- function(p) {
  type <- extract_RcppR6_template_type(p, "Parameters")
  scm <- SCM(type)(p)
  n_spp <- length(p$strategies)

  lai_error <- rep(list(NULL), n_spp)
  while (!scm$complete) {
    added <- scm$run_next()
    for (idx in added) {
      lai_error[[idx]] <-
        c(lai_error[[idx]], list(scm$area_leaf_error(idx)))
    }
  }

  lai_error <- lapply(lai_error, function(x) rbind_list(pad_matrix(x)))
  seed_rain_error <- scm$seed_rain_error
  f <- function(m) {
    suppressWarnings(apply(m, 2, max, na.rm=TRUE))
  }
  total <- lapply(seq_len(n_spp), function(idx)
                  f(rbind(lai_error[[idx]], seed_rain_error[[idx]])))

  list(seed_rain=scm$seed_rains,
       err=list(lai=lai_error, seed_rain=seed_rain_error, total=total),
       ode_times=scm$ode_times)
}

##' @rdname Hyperparameter functions
##' @param type Either \code{"FF16"} or \code{"FF16r"}.
##' @export
make_hyperpar <- function(type) {
  switch(type,
         FF16=make_FF16_hyperpar,
         FF16r=make_FF16_hyperpar,
         stop("Unknown type ", type))
}

##' @rdname Hyperparameter functions
##' @param type Either \code{"FF16"} or \code{"FF16r"}.
##' @export
hyperpar <- function(type) {
  switch(type,
         FF16=FF16_hyperpar,
         FF16r=FF16_hyperpar,
         stop("Unknown type ", type))
}

##' Helper function for creating parameter objects suitable for an
##' assembly.
##' @title Helper function for creating parameter objects
##' @param ... Named set of parameters
##' @param pars A list of parameters
##' @export
assembly_parameters <- function(..., pars=NULL) {
  p <- plant::scm_base_parameters()

  ## These are nice to have:
  p$control$equilibrium_solver_name <- "hybrid"
  p$control$equilibrium_nsteps <- 60

  if (is.null(pars)) {
    pars <- list(...)
  } else if (length(list(...)) > 0L) {
    stop("Do not provide both ... and pars")
  }

  if (length(pars) > 0L) {
    assert_named_if_not_empty(pars)

    excl <- c("control", "strategy_default", "hyperpar")
    pos <- setdiff(c(names(formals(make_FF16_hyperpar)),
                     names(p),
                     names(p$control),
                     names(p$strategy_default)),
                   excl)
    unk <- setdiff(names(pars), pos)
    if (length(unk) > 0L) {
      stop("Unknown parameters: ", paste(unk, collapse=", "))
    }

    nms_hyper <- intersect(names(pars), names(formals(make_FF16_hyperpar)))
    p$hyperpar <- do.call("make_FF16_hyperpar", pars[nms_hyper])
    p                  <- modify_list(p,                  pars)
    p$control          <- modify_list(p$control,          pars)
    p$strategy_default <- modify_list(p$strategy_default, pars)
  }
  p
}

scm_to_internals <- function(obj, use_environment=TRUE) {
  dat <- lapply(seq_along(obj$time), function(i)
    patch_to_internals(scm_patch(i, obj), use_environment))
  f <- function(i) {
    aperm(pad_list_to_array(lapply(dat, function(x) t(x[[i]]))), c(1, 3, 2))
  }
  lapply(seq_along(dat[[1]]), f)
}

patch_to_internals <- function(x, use_environment=TRUE) {
  env <- if (use_environment) x$environment else NULL
  lapply(x$species, species_to_internals, env)
}

species_to_internals <- function(sp, environment=NULL) {
  sp_pp <- lapply(sp$cohorts, function(x)
    plant_to_plant_plus(x$plant, environment))
  ints <- do.call("rbind", lapply(sp_pp, function(x) unlist(x$internals)))
  cbind(ints,
        log_density=sp$log_densities,
        seeds_survival_weighted=sp$seeds)
}

##' Create a function that allows integrating aggregate properties of
##' the SCM system.
##'
##' The workflow here is to run an SCM to create an SCM by running
##' \code{run_scm}, or a set of data from \code{run_scm_collect} and
##' then reconstitute all the intermediate bits of data so that an any
##' variable that \code{PlantPlus} tracks can be integrated out.
##' Because the pre-processing step is reasonably slow, this function
##' returns a function that takes a variable name and integrates it.
##'
##' @title Integrate SCM variables
##' @param scm An object from \code{run_scm} or \code{run_scm_collect}
##' @export
make_scm_integrate <- function(obj) {
  ## TODO: This needs to be made to work with the output of
  ## run_scm_collect, which means that it could possibly work with the
  ## output over time, which would be cool; that might help with some
  ## of the stuff from the "emergent" vignette.
  if (inherits(obj, "SCM")) {
    internals <- patch_to_internals(obj$patch)
    n <- length(internals)
    sched <- obj$cohort_schedule
    a <- lapply(seq_len(n), sched$times)
    pa <- lapply(a, obj$patch$environment$disturbance_regime$density)
  } else {
    internals <- patch_to_internals(scm_patch(length(obj$time), obj))
    n <- length(internals)
    a <- obj$p$cohort_schedule_times
    pa <- lapply(a, Disturbance(obj$p$disturbance_mean_interval)$density)
  }

  if (n == 0L) {
    stop("This just isn't going to work out (plant internals are empty)")
  }

  pos <- colnames(internals[[1]])

  f1 <- function(i, name, error=FALSE) {
    x <- a[[i]]
    y <- pa[[i]] * internals[[i]][, name]
    total <- trapezium(x, y)
    if (error) {
      local_error_integration(x, y, total)
    } else {
      total
    }
  }
  function(name, error=FALSE) {
    if (!(name %in% pos)) {
      stop("Unknown variable: ", name)
    }
    vnapply(seq_len(n), f1, name, error)
  }
}
