## ---
## title: "plant: A package for modelling forest trait ecology & evolution: _Patch-level dynamics_"
## ---

## The aim here is to use plant to investigate dynamics of plants
## within a single patch.
library(plant)

p0 <- scm_base_parameters("FF16")
p0$control$equilibrium_nsteps <- 30
p0$control$equilibrium_solver_name <- "hybrid"
p0$disturbance_mean_interval <- 30.0

## First, with a single species:
p1 <- expand_parameters(trait_matrix(0.0825, "lma"), p0, FALSE)

## Run everything out to equilibrium:
p1_eq <- equilibrium_seed_rain(p1)

## This collects information about the state of the system at every
## ODE step:
data1 <- run_scm_collect(p1_eq)

## Entries are:
##   * time: time of the step
##   * species: a list with the state of each species (in this case
##     there is only 1).  The contents is a 3d array of variable /
##     time step / cohort
##   * light_env: a list with the points of the light environment
##     spline (height / canopy openness)
##   * seed_rain: output seed rain
##   * p: input parameters
names(data1)

t <- data1$time
h <- data1$species[[1]]["height", , ]

## h here is two of the three dimensions of the state array; each row
## is a *time* and each column is a *cohort*.
dim(h)

## So the ith row is the height of all extant cohorts in the patch
## (some are NA if they have not been introduced yet)
h[150, ]

## And the jth column is the heights of a particular cohort, NA if
## that time is before it was introduced:
h[, 60]

## With a single species there is always one more time step than
## cohort introduction, with the last two steps having the same number
## of cohorts.  With multiple species there can be more time steps
## than cohort introductions, as we'll record data every cohort
## introduction for *either* species, which will be refined to
## different schedules.

## Individual plants increase in height with respect to time, but
## because they are competing the growth rate depends on the amount of
## shading above them, in addition to size dependent growth rate (see
## vignette:plant)
matplot(t, h, lty=1, col=make_transparent("black", 0.25), type="l",
        las=1, xlab="Time (years)", ylab="Height (m)")

## The light environment is stored over each time step:
xlim <- c(0, 1.1)
ylim <- range(data1$light_env[[length(data1$light_env)]][, "height"])
plot(NA, xlim=xlim, ylim=ylim, las=1,
     xlab="Canopy openness", ylab="Height (m)")
for (i in data1$light_env) {
  lines(i[, "canopy_openness"], i[, "height"], col="grey")
}

blues <- c("#DEEBF7", "#C6DBEF", "#9ECAE1", "#6BAED6",
           "#4292C6", "#2171B5", "#08519C", "#08306B")
times <- c(5, 10, 20, 40, data1$time[[length(data1$time)]])
cols <- colorRampPalette(blues[-(1:2)])(length(times))
for (i in seq_along(times)) {
  x <- data1$light_env[[which.min(abs(times[[i]] - data1$time))]]
  lines(x[, "canopy_openness"], x[, "height"], col=cols[[i]])
  y <- x[nrow(x), "height"]
  points(1, y, pch=19, col=cols[[i]])
  text(1 + strwidth("x"), y, paste(round(times[[i]]), "years"),
       adj=c(0, 0))
}

## The amount of light at the ground level is perhaps the most
## relevant metric:
y <- sapply(data1$light_env, function(x) x[1, "canopy_openness"])
plot(data1$time, y, type="l", las=1,
     ylim=c(0, 1), xlab="Time (years)", ylab="Canopy openness")

## The waves here are due to rounds of recruitment and self thinning.
## Mortality is not instantaneous, so species recruit to a density
## that generates a canopy that they cannot survive under.

## Leaf area index is the driver that controls the canopy openness
## (via the light extinction coefficient `p1$k_I`, following
## exponential extinction).  This is not returned by
## `run_scm_collect` so instead we need to rebuild patches using
## `scm_patch`.
patches1 <- lapply(seq_along(data1$time), scm_patch, data1)

## Each element of the resulting list is a Patch object, the same as
## was observed when running the model.
lai1 <- sapply(patches1, function(x) x$area_leaf_above(0.0))
plot(data1$time, lai1, type="l", las=1, xlab="Time (years)",
     ylab="Leaf area index")

## If multiple species are grown at once, they compete with one
## another.  This adds a second species -- with a higher lma value
## than the first species -- to the population and runs them until
## they reach equilibrium seed rain at the metapopulation level.
p2 <- expand_parameters(trait_matrix(0.2625, "lma"), p1, FALSE)
p2_eq <- equilibrium_seed_rain(p2)

## Then collect the patch-level dynamics:
data2 <- run_scm_collect(p2_eq)

t2 <- data2$time
h1 <- data2$species[[1]]["height", , ]
h2 <- data2$species[[2]]["height", , ]

cols <- c("#e34a33", "#045a8d")
matplot(t2, h1, lty=1, col=make_transparent(cols[[1]], .25), type="l",
        las=1, xlab="Time (years)", ylab="Height (m)")
matlines(t2, h2, lty=1, col=make_transparent(cols[[2]], .25))

## Alternatively we can compare the growth of species 1 by itself or
## with another species:
matplot(t, h, lty=1, col=make_transparent("black", .25), type="l",
        las=1, xlab="Time (years)", ylab="Height (m)")
matlines(t2, h1, lty=1, col=make_transparent(cols[[1]], .25))

## This shows that the additional species does not affect the growth
## of the *initial* wave of cohorts (because the second species is
## growing more slowly and is shorter than the first species), but
## subsequent waves are slowed or eliminated.

## The dynamics are easier to see when coded by cohort density (some
## of the lines here represent cohorts at close to zero density).

## Relativise the log densities onto (-4, max)
d1 <- data2$species[[1]]["log_density", , ]
d2 <- data2$species[[2]]["log_density", , ]
rel <- function(x, xmin) {
  x[x < xmin] <- xmin
  xmax <- max(x, na.rm=TRUE)
  (x - xmin) / (xmax - xmin)
}
rd1 <- rel(d1, -4)
rd2 <- rel(d2, -4)

## R doesn't seem to offer a way to plot lines that vary in colour, so
## this is quite roundabout using `segments`, shaded by the density at
## the first part of the line segment:
n <- length(t2)
x <- matrix(rep(t2, ncol(h1)), nrow(h1))
col1 <- matrix(make_transparent(cols[[1]], rd1), nrow(d1))
col2 <- matrix(make_transparent(cols[[2]], rd2), nrow(d2))
plot(NA, xlim=range(t2), ylim=range(h1, na.rm=TRUE),
     las=1, xlab="Time (years)", ylab="Cohort height (m)")
segments(x[-1, ], h2[-1, ], x[-n, ], h2[-n, ], col=col2[-n, ], lend="butt")
segments(x[-1, ], h1[-1, ], x[-n, ], h1[-n, ], col=col1[-n, ], lend="butt")

## Then, the total leaf area:

## Each element of the resulting list is a Patch object, the same as
## was observed when running the model.
patches2 <- lapply(seq_along(data2$time), scm_patch, data2)
lai2 <- sapply(patches2, function(x) x$area_leaf_above(0.0))

lai2_1 <- sapply(patches2, function(x) x$species[[1]]$area_leaf_above(0.0))
lai2_2 <- sapply(patches2, function(x) x$species[[2]]$area_leaf_above(0.0))

plot(t2, lai2, type="l", las=1, lty=2,
     xlab="Time (years)", ylab="Leaf area index")
lines(t2, lai2_1, col=cols[[1]])
lines(t2, lai2_2, col=cols[[2]])

## To find the average value over the metapopulation we weight by patch abundance:
metapopulation <- function(x){
  plant:::trapezium(t2, x*data2$patch_density)
}

lai2_av <- metapopulation(lai2)
lai2_1_av <- metapopulation(lai2_1)
lai2_2_av <- metapopulation(lai2_2)

plot(t2, lai2, type="l", las=1, lty=2,
     xlab="Time (years)", ylab="Leaf area index")
lines(t2, lai2_1, col=cols[[1]])
lines(t2, lai2_2, col=cols[[2]])

axis(4, at=lai2_av,   tck=0.1, lty = 2, labels=NA)
axis(4, at=lai2_1_av, tck=0.1, col.ticks=cols[[1]], labels=NA)
axis(4, at=lai2_2_av, tck=0.1, col.ticks=cols[[1]], labels=NA)
axis(1, at=108, labels="Av")
